library playfield;

import 'package:built_value/built_value.dart';
import 'package:wui_builder/wui_builder.dart';
import 'package:built_collection/built_collection.dart';
import 'package:wui_builder/vsvg.dart';

import 'models/tetromino.dart';
import 'location.dart';

part 'playfield.g.dart';

abstract class PlayFieldProps
    implements Built<PlayFieldProps, PlayFieldPropsBuilder> {
  PlayFieldProps._();
  factory PlayFieldProps(BuiltList<BuiltList<Tetromino>> board) =>
      new _$PlayFieldProps._(board: board);

  BuiltList<BuiltList<Tetromino>> get board;
}

class PlayField extends Component<PlayFieldProps, Null> {
  PlayField(Null props) : super(props);

  @override
  render() => new VRectElement()
    ..attributes = {
      'height': '1000',
      'width': '250',
      'x': '375',
      'y': '0',
      'fill': '#000000',
    }
    ..children = _locations();

  Iterable<VNode> _locations() {
    final locations = new List<Location>();
    for (int i = 0; i < props.board.length; i++) {
      for (int j = 0; j < props.board[i].length; j++) {
        if (props.board[i][j] != null) locations.add(_location(i, j));
      }
    }
    return locations;
  }

  Location _location(int i, int j) => new Location(
        new LocationProps((b) => b
          ..tetromino = props.board[i][j].toBuilder()
          ..x = i
          ..y = j),
      );
}
